<?php
/**
 * @package rgraph Joomla! Plugin
 * @version Version 1.2.5
 * @author  Joachim Schmidt - joachim.schmidt@jschmidt-systemberatung.de
 * @copyright (C) 2016 Joachim Schmidt. All rights reserved.
 * @license GNU/GPLv3 http://www.gnu.org/licenses/gpl-3.0.html
 *
 * change activity:
 *  01.10.2018: changed to namespaced classes (Joomla 3.x.API)
 *  08.08.2020: changed to support Jommla V4
 *  01.03.2021: added modal popup window
 *  04.03.2021: changed to support Jommla V4 with bootstrap Version 5
 *  31.01.2024: added support for horizontalLines
 */
namespace rgraph\plugin;
// no direct access
defined('_JEXEC') or die('Restricted access');

use Joomla\CMS\Language\Text;
use Joomla\CMS\Factory;
use Joomla\CMS\Date\Date;
use rgraph\lib\plg_rgraph_chart;
use rgraph\lib\plg_rgraph_options;
use rgraph\lib\EvalMath;
use rgraph\lib\SimpleLinearRegression;
use DateTime;
use DateTimeZone;

class plgContentrgraphHelper
{

    var $options;

    function convert2Array ($param_line)
    {
        $matches = array();
        preg_match_all('/(\w+)(\s*=\s*\".*?\")/s', $param_line, $matches);
        $count = count($matches[1]);
        $parms = array();

        for ($i = 0; $i < $count; $i ++)
        {
            $value = ltrim($matches[2][$i], " \n\r\t=");
            $value = trim($value, '"');
            $parm = $matches[1][$i];
            $parms[$parm] = $value;
        }
        return $parms;
    }

    function getParam (&$parms, $type, $attribute = null)
    {
        if (array_key_exists($type, $parms))
        {
            $result = $parms[$type];
            unset($parms[$type]);
            return $result;
        }
        else
            return $attribute;
    }

    function buildModalhtml ($id, $content, $link, $title = null, $button = false, $width = 500, $height = 350)
    {
        $header_style = "";
        $data = "data-";

        $div = $id . '_modal';
        $min_width = $width + 10;

        if ($button)
            $html = '<!-- Trigger the modal with a button -->
<button type="button" class="btn btn-default" ' . $data . 'toggle="modal" ' . $data . 'target="#' . $div . '">' . $link . '</button>';
        else
            $html = '<!-- Trigger the modal with a link -->
<a ' . $data . 'toggle="modal" href="#' . $div . '">' . $link . '</a>';

        $html .= "\n" .'<!-- Modal Window -->
<div id="' . $div . '" class="modal fade" role="document">
  <div class="modal-dialog modal-lg" style="width:' . $width . 'px; min-width:' . $min_width . 'px;">

    <!-- Modal content-->
    <div class="modal-content">
      <div class="modal-header"' . $header_style . '>
        <span class="modal-title">' . $title . '</span>
        <button type="button" class="close" ' . $data . 'dismiss="modal" title="' . Text::_('CLOSE') . '">&times;</button>
      </div>';
        if (strncmp($content, "<canvas", 7) == 0)
            $html .= '<div id="' . $div . '_body" class="modal-body">' . $content . '</div>';
        else
        {
            $html .= '<div id="' . $div . '_body" class="modal-body">';
            $html .= '    <iframe src="' . $content . '" width="' . $width . 'px" height="' . $height . 'px" frameborder="0" allowtransparency="true" scrolling="no"></iframe>';
            $html .= '</div>';
        }
        $html .= '</div>
  </div>
</div>';

        return $html;
    }

    function getChartProperties ($id, $plugin_parms, $param_line, &$row)
    {
        $result = array();
        $error_string = "";

        $debug = $this->getParam($param_line, 'debug');
        
        $suppress_error = $this->getParam($param_line, 'suppress_error', $this->getParam($plugin_parms, 'suppress_error', false));
        $width = $this->getParam($param_line, 'width', $this->getParam($plugin_parms, 'width', 400));
        $height = $this->getParam($param_line, 'height', $this->getParam($plugin_parms, 'height', 300));
        $responsive = $this->getParam($param_line, 'responsive', $this->getParam($plugin_parms, 'responsive', ""));

        $chart_alignment = $this->getParam($param_line, 'chart_alignment', $this->getParam($plugin_parms, 'chart_alignment', 'none'));
        $canvas_class = $this->getParam($param_line, 'canvas_class', $this->getParam($plugin_parms, 'canvas_class', ''));
        $number_format = $this->getParam($param_line, 'number_format', $this->getParam($plugin_parms, 'number_format', 'c'));
        $tooltipsEffect = $this->getParam($param_line, 'tooltipsEffect', $this->getParam($plugin_parms, 'tooltipsEffect', 'fade'));
        $modal_chart = $this->getParam($param_line, 'modal_chart');
        $popup_window = $this->getParam($param_line, 'popup_window');
        $create_png = $this->getParam($param_line, 'create_png');
        $title = $this->getParam($param_line, 'title');
        $subtitle = $this->getParam($param_line, 'titleSubtitle');
        if ($subtitle == "")
          $subtitle = $this->getParam($param_line, 'titleTopSubtitle');
        $type = $this->getParam($param_line, 'type');
        $draw_option = $this->getParam($param_line, 'draw_option');
        $combine_chart = $this->getParam($param_line, 'combine_chart');
        $horizontalLines = $this->getParam($param_line, 'horizontalLines');
        $drill_chart = $this->getParam($param_line, 'drill_chart');
        $drill_charts = $this->getParam($param_line, 'drill_charts');
        $drill_tips = $this->getParam($param_line, 'drill_tips');
        $tips_width = $this->getParam($param_line, 'tips_width', 150);
        $tips_height = $this->getParam($param_line, 'tips_height', 110);

        // data parms
        $data = $this->getParam($param_line, 'data');
        $bubbleData = $this->getParam($param_line, 'bubbleData');
        $max_size = $this->getParam($param_line, 'max_size');

        $min = $this->getParam($param_line, 'min');
        $max = $this->getParam($param_line, 'max');
        $value = $this->getParam($param_line, 'value');
        $csv_parm = $this->getParam($param_line, 'csv_parm');
        $csv_cols = $this->getParam($param_line, 'csv_cols');
        $sheets_parm = $this->getParam($param_line, 'sheets_parm');
        $sheets_cols = $this->getParam($param_line, 'sheets_cols');
        $file = $this->getParam($param_line, 'file');
        $url = $this->getParam($param_line, 'url');
        $sql = $this->getParam($param_line, 'sql');
        $sql_labels = $this->getParam($param_line, 'sql_labels');
        $scatter_data = $this->getParam($param_line, 'scatter_data');
        $x_step = $this->getParam($param_line, 'x_step');
        $log_scale = $this->getParam($param_line, 'log_scale');
        $gradient_colors = $this->getParam($param_line, 'gradient_colors');

        $formula = $this->getParam($param_line, 'formula');
        $x_min = $this->getParam($param_line, 'x_min');
        $x_max = $this->getParam($param_line, 'x_max');
        $x_interval = $this->getParam($param_line, 'x_interval');
        $allow_formula = $this->getParam($plugin_parms, 'allow_formula', '0');

        $calculate_regression = $this->getParam($param_line, 'calculate_regression');
        $show_regression_formula = $this->getParam($param_line, 'show_regression_formula');
        $rearrangedata = $this->getParam($param_line, 'rearrangedata');
        $reverse_csv = $this->getParam($param_line, 'reverse_csv');

        // optional parms
        $alpha = $this->getParam($param_line, 'alpha');
        $chart_name = $this->getParam($param_line, 'chart_name');
        $truncate_label = $this->getParam($param_line, 'truncate_label');
        $tooltip = $this->getParam($param_line, 'tooltip');
        $tooltips = $this->getParam($param_line, 'tooltips');
        $chart_script = $this->getParam($param_line, 'chart_script');
        $event_script = $this->getParam($param_line, 'event_script');
        $allow_script = $this->getParam($plugin_parms, 'allow_script', '0');
        $allow_user_table = $this->getParam($plugin_parms, 'allow_user_table', '0');
        $dbhost = $this->getParam($plugin_parms, 'dbhost', 'localhost');
        $dbname = $this->getParam($plugin_parms, 'dbname', '');
        $dbuser = $this->getParam($plugin_parms, 'dbuser', '');
        $dbpassword = $this->getParam($plugin_parms, 'dbpassword', '');

        /*
         * ------------------------------------------- *
         * process parameters for rgraph object        *
         * ------------------------------------------- *
         */
        $result['debug'] = false;
        $result['suppress_error'] = false;
        
        if ($suppress_error != 0 && $suppress_error != 'false')
          $result['suppress_error'] = true;
        
        if ($id == '')
            $id = "chart_" . rand(1, 9999);
                    
        // chart object
        $chart = new plg_rgraph_chart($id, null, $type);
        $options = new plg_rgraph_options();
        $noaxis_charts = array(
                "check",
                "Pie",
                "Fuel",
                "Funnel",
                "Gauge",
                "Meter",
                "Odometer",
                "Radar",
                "Rose",
                "RScatter"
        );

        $notooltip_charts = array(
                "check",
                "Gauge",
                "Meter",
                "Odometer"
        );

        if (array_search($type, $noaxis_charts) != false)
        {
            foreach ($plugin_parms as $key => $val)
            {
                if (substr($key, 0, 1) == "x" || substr($key, 0, 1) == "y")
                    unset($plugin_parms[$key]);
            }
        }

        if (array_search($type, $notooltip_charts) == false)
        {
            if (empty($tooltipsEffect))
               $options->set_option('tooltipsEffect', 'fade');
            elseif ($tooltipsEffect == 'slide' || $tooltipsEffect == 'fade' )
               $options->set_option('tooltipsEffect', $tooltipsEffect);
            else
               $options->set_option('tooltipsEffect', 'none');
        }
            
        if (array_search($type, $notooltip_charts) != false)
        {
            foreach ($plugin_parms as $key => $val)
            {
                if (substr($key, 0, 4) == "tool" || strcmp($key, "colors") == 0)
                {
                    unset($plugin_parms[$key]);
                    unset($options->colors);
                    unset($plugin_parms['backgroundImage']);
                }
            }
        }

        if ($debug == "1" || $debug == "true")
        {
            $debug = true;
            $result['debug'] = true;
            $options->set_option("debug", true);
            $options->set_debuginfo("check setting:");
        }

        $options->set_optionsviaArray($plugin_parms);
        $options->set_optionsviaArray($param_line);

        if (! empty($chart_name))
            $result['chart_name'] = $chart_name;
        else
            $result['chart_name'] = "chart" . $id;

        if ($responsive != "")
        {
            if (!empty($draw_option))
             $draw_option = trim($draw_option, ";");
            $parm = explode(",", $responsive);

            if (count($parm) == 3)
            {
                array_push($parm, $width, $height);
                $responsive = $options->getResponsiveFunction($parm);
            }
            else
                $responsive = ";";
        }
           
        if ($draw_option != "")
        {
            $trans = array(
                    "[" => "{",
                    "]" => "}"
            );
            $draw_option = strtr($draw_option, $trans);
            $result['draw_option'] = "." . $draw_option . $responsive;
        }
        else
            $result['draw_option'] = ".draw()" . $responsive . ";";

        if ($sql_labels !== "0" && $sql_labels !== "false")
            $sql_labels = "1";

        if ($canvas_class != "")
            $result['canvas_class'] = $canvas_class;

        if ($chart_alignment != "")
        {
            if ($chart_alignment == "right")
                $result['chart_alignment'] = "float: right;display: inline-block;";
            elseif ($chart_alignment == "left")
                $result['chart_alignment'] = "float: left;display: inline-block;";
            elseif ($chart_alignment == "center")
                $result['chart_alignment'] = "margin-left:auto; margin-right:auto; display: inline-block;";
            else
                $result['chart_alignment'] = "display: inline-block;";
        }
        else
            $result['chart_alignment'] = "display: inline-block;";

        if ($width != "")
        {
            if (is_numeric($width))
                $result['width'] = (integer) $width;
            else
                $result['width'] = 400;
        }

        if ($height != "")
        {
            if (is_numeric($height))
                $result['height'] = (integer) $height;
            else
                $result['height'] = 300;
        }

        if ($combine_chart == "1" || $combine_chart == "true")
        {
            $result['combine_chart'] = true;
            $options->set_option("backgroundImage", null, true);
            $options->set_option("backgroundColor", "", true);
            $options->set_option("backgroundGrid", false, true);
        }
        else
            $result['combine_chart'] = false;

        if ($drill_chart != "")
        {
           
            $tmp = explode(",", $drill_chart);
            $result['drill_chart'] = $tmp[0];
            if (! empty($tmp[1]) && $tmp[1] == "1" && empty($options->contextmenu))
            {
                $item = Text::_('SHOW_MAIN_CHART');
                $options->set_option("contextmenu", array(
                        array(
                                $item,
                                $id . "_main"
                        )
                ));
            }
        }
        else
            $result['drill_chart'] = "";

        if ($drill_tips != "")
        {
            $chart_name = $result['chart_name'];
            $result['drill_tips'] = explode(",", $drill_tips);
            $options->set_option("drill_tips", $result['drill_tips'], true);
            $tooltips = "JAVA_VARfunction (index) {  var label = " . $chart_name . ".get('labels')[index]; return QQ<b>QQ + label + QQ</b><br><canvas id='tooltip_canvas' width='" . $tips_width . "' height='" . $tips_height . "'></canvas>QQ }JAVA_VAR";
            $options->set_option("tooltips", $tooltips, true);
        }
        else
            $result['drill_tips'] = "";

        if ($drill_charts != "")
        {
            $result['drill_charts'] = explode(",", $drill_charts);
            $options->set_option("drill_charts", $result['drill_charts'], true);
        }
        else
            $result['drill_charts'] = "";

        if ($title != '')
        {
            $title = str_replace("<br>", "\n", $title);
            $title = str_replace("<br />", "\n", $title);
            if (strpos($title, "#varinfo#") !== false)
                $error_string = $this->replaceVar($row, $title);
            elseif (strpos($title, "#") !== false)
                $title = $this->replaceVar($row, $title);
            $result['title'] = $title;
            $options->set_option("title", $title, true);
        }
        
        if ($subtitle != '')
        {
            $subtitle = str_replace("<br>", "\n", $subtitle);
            $subtitle = str_replace("<br />", "\n", $subtitle);
            if (strpos($subtitle, "#varinfo#") !== false)
                $error_string = $this->replaceVar($row, $subtitle);
                elseif (strpos($subtitle, "#") !== false)
                $subtitle = $this->replaceVar($row, $subtitle);
                if ($type == 'Gauge')
                {
                  $result['titleTopSubtitle'] = $subtitle;
                  $options->set_option("titleTopSubtitle", $subtitle, true);
            }
            else
            {
                $result['titleSubtitle'] = $subtitle;
                $options->set_option("titleSubtitle", $subtitle, true);
            }
        }


        if ($modal_chart != '')
        {
            $parm = explode(",", $modal_chart);
            $modal_chart = array();

            if (count($parm) == 1)
            {
                $modal_chart[0] = $parm[0];
                $modal_chart[1] = false;
            }
            else
            {
                $modal_chart[0] = $parm[0];
                $modal_chart[1] = (boolean) trim($parm[1]);
            }
            $options->set_option("textAccessible", false, true);
            $result['modal'] = $modal_chart;
        }

        if ($popup_window != '')
        {
            $parm = explode(",", $popup_window);
            $modal_chart = array();

            if (count($parm) == 1)
            {
                $modal_chart[0] = $parm[0];
                $modal_chart[1] = false;
            }
            else
            {
                $modal_chart[0] = $parm[0];
                $modal_chart[1] = (boolean) trim($parm[1]);
            }
            $result['error'] = null;
            if ($url == '')
                $result['error'] = Text::_('NOPOPUP_URL_SPECIFIED');
            else
                $result['url'] = $url;
            $result['popup'] = $modal_chart;

            return $result;
        }
        else
            $result['popup'] = '';

        if ($create_png != '1' && $create_png != 'true')
            $create_png = false;
        else
        {
            $allow_script = true;
            $chart_script = 'media/plg_content_rgraph/js/showpng.js';
        }
        
        if ($gradient_colors)
            $options->set_gradientColors ();
            
        if ($alpha != '')
        {
            if (is_numeric($alpha))
            {
                if ($alpha <= 1 && $alpha > 0)
                    $options->setColorOpacity($alpha);
            }
            else
                $error_string .= Text::_('INVALID_VALUE') . " alpha ($alpha) " . Text::_('DATA_NOT_NUMERIC');
        }

        if ($x_step != '' && $type != "Pie")
        {
            if (! is_numeric($x_step))
                $error_string .= Text::_('INVALID_VALUE') . " x_step ($x_step) " . Text::_('DATA_NOT_NUMERIC');
        }

        if ($truncate_label != "")
        {
            $truncate_parms = explode(",", $truncate_label);
            if (! isset($truncate_parms[1]))
                $truncate_parms[1] = "";
        }
        else
            $truncate_parms = null;

        if ($min != "")
        {
            settype($min, "float");
            $chart->set_min($min);
        }

        if ($max != "")
        {
            settype($max, "float");
            $chart->set_max($max);
        }

        if ($value != "")
        {
            settype($value, "float");
            $chart->set_value((float) $value);
        }

        if ($max_size != "")
        {
            settype($max_size, "integer");
            $result['max_size'] = $max_size;
        }
        else
        {
            $result['max_size'] = 50;
            $max_size = 50;
        }

        if ($calculate_regression != '' && $data != "formula")
        {
            if ($show_regression_formula != '' && $show_regression_formula != '0')
                $show_regression_formula = "1";
        }

        if ($formula != '')
        {
            if ($allow_formula != "1" && $allow_formula != "true")
                $error_string .= Text::_('FORMULA_NOT_ALLOWED');
        }

        if ($number_format != '')
            $options->numberFormat($number_format, $type);

        if ($horizontalLines)
        {
            $tmp = explode(",", $horizontalLines);

            $line_label = str_replace('#val#', '%{value}', $tmp[0]);
                      
            $value = 'average';
            if (isset($tmp[1]))
            {
                if (is_numeric($tmp[1]))
                    $value = intval($tmp[1]);
            }
                
            $position = 'top right';
            if (isset($tmp[2]))
            {
                if (trim($tmp[2]) != '')
                    $position = trim($tmp[2]);
            }
            
            $color = $options->get_option('textColor');
            if (isset($tmp[3]))
            {
                if (trim($tmp[3]) != '')
                    $color = trim($tmp[3]);
            }

            $line_width = $options->get_option('linewidth');
            if (isset($tmp[4]))
            {
                if (trim($tmp[4]) != '')
                    $line_width = floatval($tmp[4]);
            }

            $dashed = true;
            if (isset($tmp[5]) )
            {
               if (intval($tmp[5]))
                 $dashed = true;
               else
                 $dashed = false;
            }

            $size = $options->get_option('textSize');

            $options->set_option("horizontalLines",
                    array(
                            array(
                                    'value' => $value,
                                    'color' => $color,
                                    'dashed' => $dashed,
                                    'linewidth' => $line_width,
                                    'label' => $line_label,
                                    'labelPosition' => $position,
                                    'labelSize' => $size,
                                    'labelValueDecimals' => $options->get_option('scaleDecimals'),
                                    'labelValueThousand' => $options->get_option('scaleThousand'),
                                    'labelValuePoint' => $options->get_option('scalePoint')
                            )));
            }

        if ($data == "csv" && $csv_parm == '')
            $error_string .= Text::_('NOCSV_SPECIFIED');

        if ($data == "sheets" && $sheets_parm == '')
            $error_string .= Text::_('NOSHEETS_SPECIFIED');

        if ($data == "file" && $file == '')
            $error_string .= Text::_('NOFILE_SPECIFIED');

        if ($data == "url" && $url == '')
            $error_string .= Text::_('NOURL_SPECIFIED');

        if ($data == "database" && $sql == "")
            $error_string .= Text::_('NOSQL_SPECIFIED');

        // build scripts for chart
        if ($chart_script != "" && $allow_script == "1")
        {
            $result['chart_script'] = $chart->getChartScript($chart_script);
        }
        elseif ($chart_script != "")
            $error_string .= Text::_('SCRIPT_NOT_ALLOWED');

        if ($event_script != "" && $allow_script == "1")
        {
            $result['event_script'] = $chart->getEventScript($event_script);
        }
        elseif ($event_script != "")
            $error_string .= Text::_('SCRIPT_NOT_ALLOWED');

        // retrieve data for chart
        if ($file != '' && $data == "file")
        {
            $filename = trim($file, "/");
            if (! file_exists($filename))
                $error_string .= "<br />&nbsp; - &quot;$file&quot; - " . Text::_('FILE_NOT_FOUND');
            else
            {
                $data = $chart->getDataFromFile($filename);
                if ($data === null)
                    $error_string .= "<br />&nbsp; - &quot;$file&quot; - " . Text::_('SOURCE_NO_DATA') . $url . "&quot";
                else
                {
                    $tmp_data = array();
                    $keys = array_keys($data);
                    $labels = $data[$keys[0]];
                    for ($i = 1; $i < count($keys); $i ++)
                    {
                        array_push($tmp_data, $data[$keys[$i]]);
                    }
                    $data = $tmp_data;
                }
            }
        }

        elseif ($csv_parm != '' && $data == "csv")
        {
            $header_lines = 0;
            $sepchar = ";";
            $cols = null;

            $parms = explode(",", $csv_parm, 3);

            $csv_file = $parms[0];
            if (isset($parms[1]))
                $header_lines = $parms[1];
            if (isset($parms[2]))
                $sepchar = $parms[2];

            if ($csv_cols != "")
            {
                $cols = explode(",", $csv_cols);
                $cols = plg_rgraph_chart::set_numeric($cols, "integer");
            }

            if ($reverse_csv == "1" || $reverse_csv == "true")
                $reverse_csv = true;
             else
                $reverse_csv = false;
                
            $csv_data = $chart->getCSVdata($csv_file, $header_lines, $sepchar, $cols, $reverse_csv);

            if ($csv_data == - 1 || $data === null)
                $error_string .= "&nbsp;" . Text::_('SOURCE_NO_DATA') . $csv_file . "&quot";
            else
            {
                if (! isset($options->labels))
                {
                    $labels = $csv_data[0];
                    array_shift($csv_data);
                }

                if (empty($csv_data))
                    $error_string .= "<br />&nbsp; - csv_cols=&quot;$csv_cols&quot; - " . Text::_('SOURCE_NO_DATA') . " from csv file";
                else
                    $data = $csv_data;
            }
        }

        elseif ($sheets_parm != '' && $data == "sheets")
        {
            $lines = 999999;
            $keep_keys = true;

            $parms = explode(",", $sheets_parm);
            $key = $parms[0];

            if (isset($parms[1]))
                $lines = (integer) $parms[1];
            else
                $lines = null;
            if (isset($parms[2]))
                if ($parms[2] == "0" || $parms[2] == "false")
                    $keep_keys = false;

            $data = $chart->getsheetsData($key, $lines, $keep_keys);
         
            if ($data === null)
                $error_string .= "<br />&nbsp; - sheets_parm =<code> $sheets_parm </code> - " . Text::_('SOURCE_NO_DATA') . " from Google sheets";
            else
            {
                $keys = array_keys($data);

                if ($sheets_cols != "")
                    $cols = explode(",", $sheets_cols);
                else
                    $cols = $keys;

                $cols = plg_rgraph_chart::set_numeric($cols, "integer");

                $labels = $data[$keys[0]];
                $tmp_data = array();
                
            
                if ($labels[0] == null)
                {
                    $labels[0] = "";
                   // $labels = array_values($labels);
                }
              
                for ($i = 1; $i < count($keys); $i ++)
                {
                    for ($j = 0; $j < count($cols); $j ++)
                    {
                        if ($cols[$j] === $keys[$i])
                        {
                            $data[$keys[$i]] = plg_rgraph_chart::set_numeric($data[$keys[$i]]);
                            array_push( $tmp_data, $data[$keys[$i]] );
                        }
                    }
                }
                $data = $tmp_data;
                if (empty($data))
                    $error_string .= "<br />&nbsp; - sheets_cols=&quot;$sheets_cols&quot; - " . Text::_('SOURCE_NO_DATA') . " from Google sheets";
            }
        }

        elseif ($url != '' && $data == "url")
        {
            $hdrs = @get_headers($url);
            if (! is_array($hdrs) ? preg_match('/^HTTP\\/\\d+\\.\\d+\\s+2\\d\\d\\s+.*$/', $hdrs[0]) : false)
                $error_string .= "<br />&nbsp; - &quot;$url&quot; - " . Text::_('URL_NOT_FOUND');
            else
            {
                $data = $chart->getDataFromUrl($url);
                if ($data === null)
                    $error_string .= "<br />&nbsp; - &quot;$url&quot; - " . Text::_('SOURCE_NO_DATA') . $url . "&quot";
                else
                {
                    $tmp_data = array();
                    $keys = array_keys($data);
                    $labels = $data[$keys[0]];
                    for ($i = 1; $i < count($keys); $i ++)
                    {
                        array_push($tmp_data, $data[$keys[$i]]);
                    }
                    $data = $tmp_data;
                    if ($data === null)
                        $error_string .= "<br />&nbsp; - &quot;$url&quot; - " . Text::_('SOURCE_NO_DATA') . " from Google sheets";
                }
            }
        }

        elseif ($data == "formula" && $error_string == "")
        {
            $lib = dirname(__FILE__) . '/lib/evalMath.php';
            require_once ($lib);

            if ($formula == null && ($allow_formula == "1" || $allow_formula == "true"))
                $error_string .= Text::_('NOFORMULA_SPECIFIED');

            if ($allow_formula == "0" || $allow_formula == "false")
                $error_string .= Text::_('FORMULA_NOT_ALLOWED');

            if ($x_interval != '')
            {
                if (! is_numeric($x_interval))
                    $error_string .= Text::_('INVALID_VALUE') . " x_interval ($x_interval) " . Text::_('DATA_NOT_NUMERIC');
            }
            if ($error_string == "")
            {
                $data = $this->getDataviaFormula($formula, $x_interval, $x_min, $x_max);
                if (isset($data['error']) || $data['y_values'] == "")
                    $error_string .= Text::_('NOFORMULA_DATA') . $data['error'];
                else
                {
                    if ($x_step == "")
                        $x_step = 1;
                    $labels = $data['x_values'];
                    $data = $data['y_values'];
                }
            }
        }

        elseif ($data == "database" && $error_string == '')
        {
            if ($dbhost != '' && $dbname != '' && $dbuser != '')
            {
                $connection = mysqli_connect($dbhost, $dbuser, $dbpassword, $dbname);
            }
            else
            {
                //$config = Factory::getConfig();
                $config = Factory::getApplication();
                $db = $config->get('db');
                $dbuser = $config->get('user');
                $dbpw = $config->get('password');
                $dbhost = $config->get('host');
                $connection = mysqli_connect($dbhost, $dbuser, $dbpw, $db);
            }

            if (strpos($sql, "#varinfo#") !== false)
                $error_string = $this->replaceVar($row, $sql);
            elseif (stripos($sql, "#") !== false)
                $sql = $this->replaceVar($row, $sql);

            // Remove html-code to cleanup user input
            $search = array(
                    '@<script[^>]*?>.*?</script>@si',
                    '@<[\/\!]*?[^<>]*?>@si',
                    '@<style[^>]*?>.*?</style>@siU',
                    '@<![\s\S]*?--[ \t\n\r]*>@'
            );
            $sql = preg_replace($search, '', $sql);
            $sql = str_replace("&nbsp;", "", $sql);
            $sql = str_replace("&lt;", "<", $sql);
            $sql = str_replace("&gt;", ">", $sql);
            $sql = str_replace("\n", "", $sql);

            $check = trim($sql, " ");

            if (stripos($check, "select") != 0)
                $error_string .= Text::_('SQL_NOTALLOWED') . $check;

            if ($allow_user_table == false)
            {
                if (stripos($check, "_users") != false)
                    $error_string .= Text::_('SQL_NOTALLOWED') . $check;
            }

            $data = $chart->getDatafromDB($connection, $sql);
            if (is_array($data))
            {

                if (isset($data[1]) && ($sql_labels !== "0" && $sql_labels !== "false"))
                {
                    $labels = $data[0];
                    $j = 1;
                }
                else
                    $j = 0;

                $tmp = array();
                for ($i = $j; $i < count($data); $i ++)
                    array_push($tmp, $data[$i]);

                $data = $tmp;
            }
            else
                $error_string .= Text::_('SQL_NO_DATA') . $sql . "&quot;";
        }

        else
        {
            if ($type != "Meter" && $type != "Odometer" && $type != "Gauge" && $type != "Fuel")
            {
                if (isset($data))
                {
                    $data = $this->getlocalData($data);
                    if (is_array($data['labels']))
                        $labels = $data['labels'];
                    $data = $data['data'];
                }
                else
                    $error_string .= Text::_('NO_DATA_TAG_FOUND');
            }
        }

        if ($calculate_regression == "1" && $type == "Scatter")
        {
            $lib = dirname(__FILE__) . '/lib/evalMath.php';
            require_once ($lib);

            $calc = $this->calculate_Regression($data[0], $data[1], $formula, $number_format, 3);
            $data = $calc['values'];
            $formula = $calc['formula'];
            $scatter_data = "1";
            if ($show_regression_formula)
                $options->set_option("key", array(
                        $formula
                ), true);
        }

        if ($error_string == "")
        {
            // got charts' data - proccess based on chart type
            if (isset($data))
            {
                if (is_array($data[0]))
                    if (count($data) == 1)
                        $data = $data[0];
            }

            if (isset($data))
            {
                if (isset($labels) && $options->get_option("labels") === null)
                    $options->set_option("labels", $labels, true);

                $chart->set_data($data);
                if ($rearrangedata == "1" || $rearrangedata == "true")
                    $chart->rearrangeData();

                if (! isset($options->ymax) && ($type == "Line" || $type == "Scatter"))
                {
                    if ($type == "Scatter" && ($scatter_data != "1" || $scatter_data != "true"))
                        $max = $this->getMaxValue($data[1]);
                    elseif ($type == "Scatter" && ($scatter_data == "1" || $scatter_data == "true"))
                        $max = $this->getMaxValue($data, true);
                    else
                        $max = $this->getMaxValue($data, false);

                    $options->set_option("ymax", $max, true);
                }

                if ($log_scale == "1" || $log_scale == "true")
                {
                    if (isset($tooltip))
                        $options->set_option('origdata', $data, true);

                    $result_log = "";
                    if (($chart->get_chartType() == "Bar" || $chart->get_chartType() == "HBar") && $rearrangedata == "" && is_array($data[0]))
                    {
                        $chart->set_data($data);
                        $data = $chart->rearrangeData(false);
                        $options->set_option("colorsSequential", false, true);
                        $rearrangedata = "0";
                    }
                    if (! is_array($data[0]))
                    {
                        $log_data = $this->calc_logs($data);
                        if ($log_data == - 1)
                            $result_log = - 1;
                    }
                    else
                    {
                        for ($i = 0; $i < count($data); $i ++)
                        {
                            $log_data[$i] = $this->calc_logs($data[$i]);
                            if ($log_data[$i] == - 1)
                                $result_log = - 1;
                        }
                    }
                    if ($result_log == - 1)
                        $error_string .= " <b>log_scaley </b>" . Text::_('INVALID_FOR_LOG_SCALE');
                    else
                    {
                        $num = null;
                        if (isset($options->yaxisLabelsCount))
                            $num = $options->yaxisLabelsCount;
                        $this->buildloglabels($type, $log_data, $number_format, $options, $num);
                        $chart->set_data($log_data);
                    }
                }
            }

            switch ($type)
            {
                case "Scatter":
                case "Rscatter":
                case "Bubble":
                    if (! empty($labels))
                        $options->set_option("xaxisLabels", $labels, true);
                    $options->set_option("xaxisLabelsOffsety", 2, true);
                    if (($scatter_data != "1" && $scatter_data != "true") && ! isset($data[1]))
                        $error_string .= "<br />&nbsp; Chart type: &quot;$type&quot; " . Text::_('NO_DATA');
                    elseif ($scatter_data != "1" && $scatter_data != "true")
                    {
                        if (! isset($data[2]))
                            $colors = null;
                        if (! isset($data[3]))
                            $tips = $data[1];

                        $labels = $data[0];
                        if ($x_step > 1)
                            $labels = $options->xlabelsteps($labels, $x_step);
                        if (! empty($labels))
                            $options->set_option("xaxisLabels", $labels, true);

                        $data = $chart->createScatterData($data[0], $data[1], $colors, $tips);
                        $chart->set_data($data);
                        $options->labels = array();
                    }
                    if ($type == "Bubble")
                    {
                        if ($bubbleData == "")
                            $error_string .= "<br />&nbsp; Chart type: &quot;$type&quot; - Parameter &quot;bubbleData&quot; " . Text::_('NO_DATA');
                        else
                        {
                            $bubble_data = explode(",", $bubbleData);
                            $max = max($bubble_data);
                            if (! isset($min))
                                $min = 0;
                            $options->set_option('bubbleData', $bubble_data);
                            $options->set_option('bubbleMax', $max);
                            $options->set_option('bubbleMin', $min);
                        }
                    }
                    break;

                case "Meter":
                case "Odometer":
                case "Gauge":
                case "Fuel":
                    if (isset($min))
                        $chart->set_min($min);
                    else
                        $chart->set_min(0);
                    if (isset($max))
                        $chart->set_max($max);
                    else
                        $error_string .= "<br />&nbsp; Chart type: &quot;$type&quot; - Parameter &quot;max&quot; - " . Text::_('NO_DATA');
                    if (isset($data[0]))
                        $value = $data[0];
                    if (isset($value))
                        $chart->set_value($value);
                    else
                        $chart->set_value(0);
                    break;

                case "Radar":
                case "Line":
                    if (! empty($labels))
                    {

                        if (is_array($truncate_parms))
                            $labels = $options->truncateLabels($labels, $truncate_parms[0], $truncate_parms[1]);
                        $xaxisLabels = $labels;
                        if ($x_step > 1 && $type == "Line")
                            $xaxisLabels = $options->xlabelsteps($labels, $x_step);

                        if ($type == "Line")
                        {
                            $options->set_option("xaxisLabels", $xaxisLabels, true);
                            // unset($options->labels);
                        }
                    }
                    if ($tooltip != "")
                        $options->set_tooltip($data, $tooltip, $type);
                    elseif ($tooltips != "")
                    {
                        $trans = array(
                                "((" => "{",
                                "))" => "}"
                        );
                        $tooltips = strtr($tooltips, $trans);
                        $options->set_option('tooltips', $tooltips, true);
                    }
                    if (isset($options->labelsAngledSpecific))
                        $options->labelsAngled = true;
                    break;

                case "Gantt":
                    if (! empty($options->labels))
                        $options->set_option("xaxisLabels", $options->labels);

                    if ($tooltip != "")
                        $options->set_tooltip($data, $tooltip, $type);
                    elseif ($tooltips != "")
                    {
                        $trans = array(
                                "((" => "{",
                                "))" => "}"
                        );
                        $tooltips = strtr($tooltips, $trans);
                        $options->set_option('tooltips', $tooltips, true);
                    }

                    if (is_array($truncate_parms))
                    {
                        $labels = $options->truncateLabels($options->labels, $truncate_parms[0], $truncate_parms[1]);
                        $options->set_option("labels", $labels, true);
                        $options->set_option("xaxisLabels", $options->labels);
                    }
                    // unset($options->labels);
                    break;

                case "Bipolar":
                case "HBar":
                    if ($type == "Bipolar")
                    {
                        if (is_array($data[0]))
                            $chart->set_left($data[0]);
                        if (is_array($data[1]))
                            $chart->set_right($data[1]);

                        $chart->data = array();
                    }

                    if (! empty($labels))
                    {

                        if (is_array($truncate_parms))
                            $yaxisLabels = $options->truncateLabels($labels, $truncate_parms[0], $truncate_parms[1]);
                        else
                            $yaxisLabels = $labels;
                        if ($x_step > 1)
                            $yaxisLabels = $options->xlabelsteps($labels, $x_step);

                        $options->set_option("labels", $labels, true);
                        $options->set_option("yaxisLabels", $yaxisLabels);
                    }

                    if ($tooltip != "")
                        $options->set_tooltip($data, $tooltip, $type);
                    elseif ($tooltips != "")
                    {
                        $trans = array(
                                "((" => "{",
                                "))" => "}"
                        );
                        $tooltips = strtr($tooltips, $trans);
                        $options->set_option('tooltips', $tooltips, true);
                    }
                    break;

                case "Pie":
                case "Funnel":
                case "Waterfall":
                    if (is_array($data[0]))
                        $chart->set_data($data[0]);
                    if ($tooltip != "")
                        $options->set_tooltip($chart->get_data(), $tooltip, $type);
                    elseif ($tooltips != "")
                    {
                        $trans = array(
                                "((" => "{",
                                "))" => "}"
                        );
                        $tooltips = strtr($tooltips, $trans);
                        $options->set_option('tooltips', $tooltips, true);
                    }
                    if (! empty($labels))
                    {
                        if (is_array($truncate_parms))
                            $labels = $options->truncateLabels($labels, $truncate_parms[0], $truncate_parms[1]);
                        if ($x_step > 1)
                            $labels = $options->xlabelsteps($labels, $x_step);
                        $options->set_option("labels", $labels, true);
                        if ($type == "Waterfall")
                            $options->set_option("xaxisLabels", $options->labels);
                    }
                    elseif ($type == "Waterfall")
                    {
                        $options->set_option("xaxisLabels", $options->labels);
                    }
                    break;

                default:
                    if (($chart->get_chartType() == "Bar" || $chart->get_chartType() == "HBar") && $rearrangedata == "" && is_array($data[0]))
                    {
                        $chart->rearrangeData();
                        $options->set_option("colorsSequential", false, true);
                    }
                    if ($chart->get_chartType() == "Bar" && empty($options->variant))
                        $options->set_option("variant", "glass", true);
                    if ($chart->get_chartType() == "Bar" && empty($options->colorsStroke))
                        $options->set_option("colorsStroke", "rgba(0, 0, 0, 0.1)", true);

                    $options->set_option("xaxisLabelsOffsety", 2, true);
                    $options->set_option("yaxisLabelsOffsetx", 2, true);

                    if (! empty($labels))
                    {
                        if (is_array($truncate_parms))
                            $labels = $options->truncateLabels($labels, $truncate_parms[0], $truncate_parms[1]);

                        $xaxisLabels = $labels;
                        if ($x_step > 1)
                            $xaxisLabels = $options->xlabelsteps($labels, $x_step);

                        $options->set_option("labels", $labels, true);
                        $options->set_option("xaxisLabels", $xaxisLabels);
                    }

                    if ($tooltip != "")
                    {
                        if ($log_scale == "1" || $log_scale == "true")
                            $options->set_tooltip($data, $tooltip);
                        else
                            $options->set_tooltip($chart->get_data(), $tooltip);
                    }
                    elseif ($tooltips != "")
                    {
                        $trans = array(
                                "((" => "{",
                                "))" => "}"
                        );
                        $tooltips = strtr($tooltips, $trans);
                        $options->set_option('tooltips', $tooltips, true);
                    }
            }
        }

        if ($options->debug != true)
            unset($options->debuginfo);
        $result['error'] = $error_string;
        $result['chart'] = $chart;
        $result['options'] = $options;
        return $result;
    }

    function calc_logs ($data)
    {
        $log_data = array();

        for ($i = 0; $i < count($data); $i ++)
        {
            if ($data[$i] >= 1)
                $log_data[$i] = log10($data[$i]);
            else
                return - 1;
        }

        return $log_data;
    }

    function buildloglabels ($type, $data, $format, $options, $num = 5)
    {
        if (isset($options->xmax) && $type == "HBar")
            $max = log10($options->xmax);
        elseif (isset($options->ymax))
            $max = log10($options->ymax);
        else
            $max = $this->getMaxValue($data, false);

        $options->yaxisScaleMax = ceil($max);

        $ymin = 0;
        $xmin = 0;

        if (isset($options->xmin) && $type == "HBar" && $options->xmin >= 1)
        {
            $options->set_option("xmin", log10($options->xmin), true);
            $xmin = $options->xmin;
        }
        elseif (isset($options->ymin) && $options->ymin >= 1)
        {
            $options->set_option("ymin", log10($options->ymin), true);
            $ymin = $options->ymin;
        }

        if ($type == "HBar")
            $val = $xmin;
        else
            $val = $ymin;
        $valy = $val;

        if ($num === null)
            $num = 4;
        $increment = ($max - $ymin) / $num;

        $pre = "";
        $post = "";
        if (isset($options->yaxisScaleUnitsPre))
            $pre = $options->yaxisScaleUnitsPre;
        if (isset($options->yaxisScaleUnitsPost))
            $post = $options->yaxisScaleUnitsPost;

        $y_labels = array();
        for ($i = 0; $i < $num + 1; $i ++)
        {
            $valy = pow(10, $val);
            $y_labels[$i] = $pre . $this->formatnumber($format, $valy, $options->scaleDecimals) . $post;
            $val += $increment;
        }

        if ($type == "HBar")
        {
            $options->set_option("xaxisLabelsSpecific", $y_labels, true);
            $options->set_option("xmax", $max, true);
        }
        else
        {
            $y_labels = array_reverse($y_labels);
            $options->set_option("yaxisLabelsSpecific", $y_labels, true);
            $options->set_option("ymax", $max, true);
        }
    }

    function getMaxValue ($data, $scatter_data = false)
    {
        if (! is_array($data))
            return null;

        $max = 0;
        if (! is_array($data[0]))
        {
            $max = max($data);
            return $max;
        }
        elseif ($scatter_data == true)
        {
            for ($i = 0; $i < count($data); $i ++)
            {
                $tmp = $data[$i][1];
                if ($tmp > $max)
                    $max = $tmp;
            }
            $max = $max * 1.05;
            return round($max, 0);
        }
        else
        {
            for ($i = 0; $i < count($data); $i ++)
            {
                $tmp = max($data[$i]);
                if ($tmp > $max)
                    $max = $tmp;
            }
            // $max = $max * 1.05;
            return $max;
        }
    }

    function getlocalData ($data)
    {
        $x_labels = array();
        $local_data = explode("/", $data);
        if (count($local_data) > 1)
        {
            $local_data[0] = trim($local_data[0]);
            $local_data[0] = str_replace("<br>", "\n", $local_data[0]);
            $local_data[0] = str_replace("&#47;", "/", $local_data[0]);
            $x_labels = explode(",", $local_data[0]);
            // $labels_found = true;
            $local_data = explode("|", $local_data[1]);
        }
        else
        {
            $local_data = explode("|", trim($data));
        }

        $tmp = array();
        for ($i = 0; $i < count($local_data); $i ++)
        {
            $tmp[$i] = explode(",", $local_data[$i]);
        }
        $local_data = plg_rgraph_chart::set_numeric($tmp);

        $result = array();
        $result['labels'] = $x_labels;
        $result['data'] = $local_data;
        return $result;
    }

    private function replaceVar (&$row, $string)
    {
        if (strpos($string, "#varinfo#") !== false)
        {
            $string = "<br /><b>" . Text::_('VAR_INFO') . "</b>";
            $vars = Factory::getApplication()->input->getArray(array());
            foreach ($vars as $key => $value)
                $string .= "<br > $key" . " = " . "$value";

            return $string;
        }

        $matches = array();
        $regex = "%#\\s*(.*?)(.*?)#%s";
        preg_match_all($regex, $string, $matches);
        for ($i = 0; $i < count($matches[1]); $i ++)
        {
            switch ($matches[2][$i])
            {
                case "date":
                    $lang = Factory::getLanguage();
                    $date = new Date('now');
                    if ($lang->getTag() == "de-DE")
                        $string = str_replace("#date#", $date->format("l, d.m.Y"), $string);
                    else
                        $string = str_replace("#date#", $date->format("l, Y-m-d"), $string);
                    break;

                case "datesql":
                    $date = new Date('now');
                    $string = str_replace("#datesql#", $date->format("Y-m-d"), $string);
                    break;

                case "time":
                    $time_zone = Factory::getApplication()->getCfg('offset');
                    $date = new DateTime('now', new DateTimeZone($time_zone));
                    $string = str_replace("#time#", $date->format("H:i"), $string);
                    break;

                case "articleid":
                    $string = str_replace("#articleid#", @$row->id, $string);
                    break;

                case "articlename":
                    $string = str_replace("#articlename#", @$row->alias, $string);
                    break;

                case "catid":
                    $string = str_replace("#catid#", @$row->catid, $string);
                    break;

                case "username":
                    $user = Factory::getUser();
                    if ($user->get('id') != 0)
                        $string = str_replace("#username#", $user->get('name'), $string);
                    else
                        $string = str_replace("#username#", Text::_('GUEST_USER'), $string);
                    break;

                case "userid":
                    $user = Factory::getUser();
                    $string = str_replace("#userid#", $user->get('id'), $string);
                    break;

                default:
                    $var = Factory::getApplication()->input->get($matches[2][$i]);
                    if ($var)
                        $string = str_replace("#" . $matches[2][$i] . "#", $var, $string);
            }
        }

        return $string;
    }

    function getDataviaFormula ($formula, $x_interval = 1, $x_min = 0, $x_max = 5, $max_iterations = 500)
    {
        $result = array();
        if (strpos($formula, "=") === false)
        {
            $user_formula = trim($formula);
            $user_formula = str_replace(",", ".", $user_formula);
        }
        else
        {
            $tmp = explode("=", $formula);
            $user_formula = trim($tmp[1]);
            $user_formula = str_replace(",", ".", $tmp[1]);
        }
        if ($user_formula)
        {
            $evalc = new EvalMath();
            $evalc->evaluate("y(x)=" . $user_formula);
        }
        else
            return null;

        if ($x_min === null)
            $x_min = 0;
        if ($x_max === null)
            $x_max = 5;
        if ($x_interval === null)
            $x_interval = 1;

        $x_values = array();
        $y_values = array();
        $max_iter = $x_max + $x_interval;

        $j = 0;

        for ($x_value = $x_min; $x_value <= $max_iter; $x_value += $x_interval)
        {
            if ($j > $max_iterations)
            {
                $result['error'] = " x_max=&quot;$x_max&quot;, x_interval=&quot;$x_interval&quot;. Maximum ($max_iterations)  " . Text::_('MAX_CALCULATIONS');
                return $result;
            }
            $j ++;
            $x = round($x_value, 5);
            $value = $evalc->e("y($x)");
            if ($evalc->last_error)
            {
                $result['error'] = " &quot;$formula&quot; Error:" . $evalc->last_error;
                return $result;
            }
            else
            {
                array_push($x_values, round($x_value, 5));
                array_push($y_values, round($value, 5));
            }
        }
        $result['x_values'] = $x_values;
        $result['y_values'] = $y_values;
        // $result['error'] = null;
        return $result;
    }

    function calculate_Regression ($x_values, $data, $user_formula = null, $number_format = "c", $precision = 3)
    {
        $R = "R²";
        $formula_out = "";
        $result = array();

        if ($user_formula)
        {
            $formula_out = $user_formula;
            if (strpos($user_formula, "=") === false)
            {
                $user_formula = trim($user_formula);
                $user_formula = str_replace(",", ".", $user_formula);
                $formula = 4;

                $calc = new EvalMath();
                $calc->evaluate("y(x)=" . $user_formula);
            }
            else
            {
                $tmp = explode("=", $user_formula);
                $user_formula = trim($tmp[1]);
                $user_formula = str_replace(",", ".", $tmp[1]);
                if ($user_formula)
                {
                    $formula = 4;
                    $calc = new EvalMath();
                    $calc->evaluate("y(x)=" . $user_formula);
                }
                else
                    $formula = false;
            }
        }
        else
        {
            $j = 0;
            $test_X = array();
            $test_Y = array();
            $test_x_values = array();
            for ($k = 0; $k < count($data); $k ++)
            {
                if (($data[$k] != 0 && $x_values[$k] != 0) && ($data[$k] !== null && $x_values[$k] !== null))
                {
                    $test_Y[$j] = log10($data[$k]);
                    $test_X[$j] = log10($x_values[$k]);
                    $test_x_values[$j] = $x_values[$k];
                    $j ++;
                }
            }

            $slr1 = new SimpleLinearRegression($x_values, $data);
            $slr2 = new SimpleLinearRegression($test_x_values, $test_Y);
            $slr3 = new SimpleLinearRegression($test_X, $test_Y);

            // test three regression formulas:
            array_multisort($x_values, $data);

            if ($slr1)
                $tst1_R = $slr1->RSquared;
            if ($slr2)
            {
                $slr2_A_value = pow(10, $slr2->YInt);
                $slr2_B_value = pow(10, $slr2->Slope);
            }
            if ($slr3)
            {
                $slr3_A_value = pow(10, $slr3->YInt);
                $slr3_B_value = $slr3->Slope;
            }

            $slr2_value = array();
            $slr3_value = array();

            foreach ($x_values as $k => $x_value)
            {
                // equation: Y = a * b^X
                $slr2_value[$k] = $slr2_A_value * pow($slr2_B_value, $x_value);
                // equation: Y = a * X^b
                $slr3_value[$k] = $slr3_A_value * pow($x_value, $slr3_B_value);
            }

            $slr2_R = SimpleLinearRegression::calcR($data, $slr2_value);
            $slr3_R = SimpleLinearRegression::calcR($data, $slr3_value);

            $tst2_R = pow($slr2_R, 2);
            $tst3_R = pow($slr3_R, 2);

            $tst2_R = $slr2_R;
            $tst3_R = $slr3_R;

            if ($tst1_R > $tst2_R && $tst1_R > $tst3_R)
                $formula = 1;

            if (is_finite($tst2_R))
            {
                if ($tst2_R > $tst1_R && $tst2_R > $tst3_R)
                    $formula = 2;
            }

            if (is_finite($tst3_R))
            {
                if ($tst3_R > $tst1_R && $tst3_R > $tst2_R)
                    $formula = 3;
            }

            if ($formula == 1)
            {
                if ($slr1->YInt >= 0)
                    $plus = "+";
                else
                    $plus = "";
                $formula_out = "y = " . $this->formatNumber($number_format, $slr1->Slope, $precision) . "x" . "$plus" . $this->formatNumber($number_format, $slr1->YInt, $precision);
            }
            $formula_out .= " ($R=" . $this->formatNumber($number_format, $slr1->RSquared, $precision) . ")";
            if ($formula == 2)
            {
                $formula_out = "y = " . $this->formatNumber($number_format, $slr2_A_value, $precision) . " * " . $this->formatNumber($number_format, $slr2_B_value, $precision) . "^x";
                $formula_out .= " ($R=" . $this->formatNumber($number_format, $slr2_R, $precision) . ")";
            }
            if ($formula == 3)
            {
                $formula_out = "y = " . $this->formatNumber($number_format, $slr3_A_value, $precision) . " * x^" . $this->formatNumber($number_format, $slr3_B_value, $precision);
                $formula_out .= " ($R=" . $this->formatNumber($number_format, $slr3_R, $precision) . ")";
            }
        }
        if ($formula)
        {
            $value = array();
            $values = array();
            $j = 0;

            foreach ($x_values as $k => $x_value)
            {
                if ($formula == 1)
                    // equation: Y = X * $Slope + $YInt
                    // $value [$k] = ($x_value * $slr1->Slope) + $slr1->YInt;
                    $value[$k] = $slr1->PredictedY[$k];
                if ($formula == 2)
                    $value[$k] = $slr2_value[$k];

                if ($formula == 3)
                    $value[$k] = $slr3_value[$k];

                if ($formula == 4)
                {
                    $x = round($x_value, 5);
                    $value[$k] = $calc->e("y($x)");

                    if ($calc->last_error)
                        return null;
                }
                if ($j == 0)
                {
                    // $values[$j] = new scatter_value($x_value, $value[$k],
                    // "");
                    $values[$j] = array(
                            round($x_value, 3),
                            round($value[$k], 3),
                            null,
                            (string) round($value[$k], $precision)
                    );
                    $j ++;
                }
                elseif ($x_values[$k] != $x_values[$k - 1])
                {
                    // $values[$j] = new scatter_value($x_value, $value[$k],
                    // "");
                    $values[$j] = array(
                            round($x_value, 3),
                            round($value[$k], 3),
                            null,
                            (string) round($value[$k], $precision)
                    );
                    $j ++;
                }
            }

            $result['values'] = $values;
            if ($formula == "4")
                $formula_out .= " ($R=" . $this->formatNumber($number_format, SimpleLinearRegression::calcR($data, $value), $precision) . ")";
            $result['formula'] = $formula_out;
            $calc = null;

            return $result;
        }
        return null;
    }

    private function formatNumber ($format, $number, $precision = 3)
    {
        switch ($format)
        {
            case "a":
                $string = number_format($number, $precision, ".", "");
                for ($i = 0; $i < $precision; $i ++)
                    $string = rtrim($string, 0);
                $string = trim($string, ".");
                break;

            case "b":
                $string = number_format($number, $precision, ",", "");
                for ($i = 0; $i < $precision; $i ++)
                    $string = rtrim($string, 0);
                $string = trim($string, ",");
                break;

            case "c":
                $string = number_format($number, $precision, ",", ".");
                for ($i = 0; $i < $precision; $i ++)
                    $string = rtrim($string, 0);
                $string = trim($string, ",");
                break;

            case "d":
                $string = number_format($number, $precision, ".", ",");
                for ($i = 0; $i < $precision; $i ++)
                    $string = rtrim($string, 0);
                $string = trim($string, ".");
                break;

            default:
                $string = number_format($number, $precision, ",", ".");
                for ($i = 0; $i < $precision; $i ++)
                    $string = rtrim($string, 0);
                $string = trim($string, ",");
                break;
        }

        return $string;
    }
}
